"""
Panopy - a Python library for handling 360° panoramic images.

Orientation is given as a quaternion in scalar-last format [x, y, z, w]. For an
equirectangular pano format this is the orientation of the center of the pano.
The coordinate system is right-handed with z as the vertical direction.
"""


from datetime import datetime
import xml.etree.ElementTree as ET

import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial.transform import Rotation
from mpl_toolkits.mplot3d import Axes3D
from PIL import Image
import cv2

class AlignmentPose():
    """
    An alignment pose is used to transform a collection of pano poses.
    """

    def __init__(self, x, y, z, orientation, name=None, validate=True):
        """
        """
        if validate is True:
            assert(isinstance(orientation, tuple) and len(orientation) == 4)

        self.x = x
        self.y = y
        self.z = z
        self.orientation = orientation
        self.name = name


class PanoPose(AlignmentPose):
    """
    A pano pose gives the position, orientation and optionally time and name of
    a pano.
    """

    def __init__(self, x, y, z, orientation, time=None, name=None, validate=True):
        """
        """
        if validate is True:
            if time is not None:
                assert(isinstance(time, datetime))

        super().__init__(x, y, z, orientation, name)

        self.time = time

    @property
    def heading(self):
        """
        Heading measured as angle from x to y axis. In equirectangular format
        this is the center of the pano. Headings are always positive to
        simplify subsequent calculations.

        See 'https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.transform.Rotation.html'
        """
        heading = Rotation.from_quat(
                self.orientation).as_euler('xyz', degrees=True)[-1]

        if heading < 0:
            heading = 360 + heading

        return heading

    @property
    def zenit(self):
        """
        Angle with the vertical
        """
        zenit = Rotation.from_quat(
                self.orientation).as_euler('xyz', degrees=True)[0]

        return zenit


    def get_direction_to_other_pose(self, other_pose, validate=True):
        """
        Angle from x to y with 'self' in the origin. Angles are always positive
        to simplify subsequent calculations.

        other_pose: instance of 'self'
        """
        if validate is True:
            assert(isinstance(self, self.__class__))

        x = other_pose.x - self.x
        y = other_pose.y - self.y
        angle = np.degrees(np.arctan2(y, x))

        if angle < 0:
            angle = 360 + angle

        return angle


class PanoPoseCollection():
    """
    A collection of pano poses with position, orientation and time.
    """

    def __init__(self, pos_xs, pos_ys, pos_zs, ori_xs, ori_ys, ori_zs,
        ori_ws, times=None, validate=True):
        """
        """
        if validate is True:
            assert(len(pos_xs) == len(pos_ys) == len(pos_zs) == len(ori_xs) ==
                len(ori_ys) == len(ori_zs) == len(ori_ws))
            if times is not None:
                assert(len(times) == len(ori_ws))

        self.pos_xs = pos_xs
        self.pos_ys = pos_ys
        self.pos_zs = pos_zs
        self.ori_xs = ori_xs
        self.ori_ys = ori_ys
        self.ori_zs = ori_zs
        self.ori_ws = ori_ws
        self.times = times

    def __len__(self):
        """
        """
        return len(self.ori_ws)

    def __getitem__(self, i):
        """
        """
        pano_pose = PanoPose(self.pos_xs[i], self.pos_ys[i], self.pos_zs[i],
            (self.ori_xs[i], self.ori_ys[i], self.ori_zs[i], self.ori_ws[i]),
            self.times[i], name=str(i), validate=False)

        return pano_pose

    def __iter__(self):
        """
        """
        for i in range(len(self)):
            yield self[i]

    @property
    def headings(self):
        """
        """
        return [p.heading for p in self]

    @property
    def zenits(self):
        """
        """
        return [p.zenit for p in self]

    @property
    def box(self):
        """
        Give bounding box of pano collection.
        """
        x_min, x_max = self.pos_xs.min(), self.pos_xs.max()
        y_min, y_max = self.pos_ys.min(), self.pos_ys.max()

        return x_min, x_max, y_min, y_max

    def transform(self, aligment_pose, validate=True):
        """
        """
        if validate is True:
            assert(isinstance(aligment_pose, AlignmentPose))

        r = Rotation.from_quat(aligment_pose.orientation)
        coordinates = np.matmul(
            r.as_matrix(), np.array([self.pos_xs, self.pos_ys, self.pos_zs]))
        pos_xs = coordinates[0] + aligment_pose.x
        pos_ys = coordinates[1] + aligment_pose.y
        pos_zs = coordinates[2] + aligment_pose.z

        # TODO: Optimise this part so operation is performed at once
        ori_xs, ori_ys, ori_zs, ori_ws = [], [], [], []
        for p in self:
            pr = r * Rotation.from_quat(p.orientation)
        
            ori_x, ori_y, ori_z, ori_w = pr.as_quat()
            ori_xs.append(ori_x)
            ori_ys.append(ori_y)
            ori_zs.append(ori_z)
            ori_ws.append(ori_w)

        ppc = PanoPoseCollection(
            pos_xs, pos_ys, pos_zs, ori_xs, ori_ys, ori_zs, ori_ws, self.times)

        return ppc

    def plot(self, headings=False, size=None):
        """
        headings: boolean (default: False) - plots headings as vectors
        with size 1.
        """
        plot_pose_collections_3D(
            [self], colors=['k'], headings=headings, size=size)


def read_leica_pano_poses_xml(filespec):
    """
    Read xml file from .e57 exported with Leica Cyclone.
    """
    root = ET.parse(filespec).getroot()

    ns = '{http://www.astm.org/COMMIT/E57/2010-e57-v1.0}'

    datetimes = []
    pos_xs, pos_ys, pos_zs = [], [], []
    ori_xs, ori_ys, ori_zs = [], [], []
    ori_ws = []

    for i, pano in enumerate(root.findall('.//{}vectorChild'.format(ns))):
        datetimes.append(datetime.fromtimestamp(float(
            pano.find('{}acquisitionDateTime'.format(ns)).find(
                '{}dateTimeValue'.format(ns)).text)))

        pose = pano.find('{}pose'.format(ns))
        if pose is None and i == 0: ## no translation and rotation given for first point
            ori_x, ori_y, ori_z = 0, 0, 0
            ori_w = 1
            pos_x, pos_y, pos_z = 0, 0, 0
        else:
            rotation = pose.find('{}rotation'.format(ns))
            ori_x = rotation.find('{}x'.format(ns)).text
            ori_y = rotation.find('{}y'.format(ns)).text
            ori_z = rotation.find('{}z'.format(ns)).text
            ori_w = rotation.find('{}w'.format(ns)).text
            if ori_x is None:
                ori_x = 0
            if ori_y is None:
                ori_y = 0
            if ori_z is None:
                ori_z = 0
            if ori_w is None:
                ori_w = 1
            
            ori_x = float(ori_x)
            ori_y = float(ori_y)
            ori_z = float(ori_z)

            translation = pose.find('{}translation'.format(ns))
            pos_x = float(translation.find('{}x'.format(ns)).text)
            pos_y = float(translation.find('{}y'.format(ns)).text)
            pos_z = float(translation.find('{}z'.format(ns)).text)

        ## Correct for Leica pano's (at least export from Cyclone) pointing
        ## towards positive y axis with center. This should be the positive
        ## x-axis so have to rotate 90 degrees.
        r = (Rotation.from_quat((ori_x, ori_y, ori_z, ori_w)) *
            Rotation.from_euler('z', 90, degrees=True))
        ori_x, ori_y, ori_z, ori_w = r.as_quat()

        ori_xs.append(ori_x)
        ori_ys.append(ori_y)
        ori_zs.append(ori_z)
        ori_ws.append(ori_w)
        pos_xs.append(pos_x)
        pos_ys.append(pos_y)
        pos_zs.append(pos_z)

    ppc = PanoPoseCollection(
        np.array(pos_xs), np.array(pos_ys), np.array(pos_zs),
        np.array(ori_xs), np.array(ori_ys), np.array(ori_zs),
        np.array(ori_ws),
        np.array(datetimes), validate=False)

    return ppc


def read_navvis_pano_poses_csv(filespec):
    """
    NavVis provides a pano-poses.csv file in each post-processed dataset. It
    holds the timestamp, position and orientation of each pano in the dataset.
    """
    with open(filespec) as f:
        data = f.read().split('\n')

    n = len(data) - 2

    datetimes = []
    pos_xs, pos_ys, pos_zs = np.zeros(n), np.zeros(n), np.zeros(n)
    ori_xs, ori_ys, ori_zs = np.zeros(n), np.zeros(n), np.zeros(n)
    ori_ws = np.zeros(n)
    for i, l in enumerate(data[1:-1]):
        r = l.split('; ')
        datetimes.append(datetime.fromtimestamp(float(r[2])))
        pos_xs[i] = r[3]
        pos_ys[i] = r[4]
        pos_zs[i] = r[5]
        ori_xs[i] = r[7]
        ori_ys[i] = r[8]
        ori_zs[i] = r[9]
        ori_ws[i] = r[6]

    ppc = PanoPoseCollection(pos_xs, pos_ys, pos_zs, ori_xs, ori_ys, ori_zs,
        ori_ws, np.array(datetimes), validate=False)

    return ppc


def read_navvis_alignment_xml(filespec):
    """
    Read xml file generated by NavVis aligment tool.
    """
    root = ET.parse(filespec).getroot()

    alignment_poses = []
    for dataset in root.findall('.//dataset'):
        name = dataset.find('name').text
        position = dataset.find('.//position')
        pos_x = float(position.find('x').text)
        pos_y = float(position.find('y').text)
        pos_z = float(position.find('z').text)
        orientation = dataset.find('.//orientation')
        ori_x = float(orientation.find('x').text)
        ori_y = float(orientation.find('y').text)
        ori_z = float(orientation.find('z').text)
        ori_w = float(orientation.find('w').text)

        alignment_poses.append(AlignmentPose(
            pos_x, pos_y, pos_z, (ori_x, ori_y, ori_z, ori_w), name))

    return alignment_poses


def plot_pose_collections(ppcs, colors=None, headings=False, size=None):
    """
    ppcs: list of PanoPoseCollection
    headings: boolean (default: False) - plots headings as vectors
    with size 1.
    """
    _, ax = plt.subplots(figsize=size)

    for i, ppc in enumerate(ppcs):
        kwargs = {}
        if colors is not None:
            kwargs['c'] = colors[i]
        ax.scatter(ppc.pos_xs, ppc.pos_ys, **kwargs)

        if headings is True:
            pc_headings = ppc.headings
            xs = np.cos(np.radians(pc_headings))
            ys = np.sin(np.radians(pc_headings))
            ax.quiver(ppc.pos_xs, ppc.pos_ys, xs, ys, color='k')

    ax.axis('equal')

    plt.show()

def plot_pose_collections_3D(ppcs, colors=None, headings=False, size=None):
    """
    ppcs: list of PanoPoseCollection
    headings: boolean (default: False) - plots headings as vectors
    with size 1.
    """
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')

    for i, ppc in enumerate(ppcs):
        kwargs = {}
        if colors is not None:
            kwargs['c'] = colors[i]
        ax.scatter(ppc.pos_xs, ppc.pos_ys, ppc.pos_zs, **kwargs)

        if headings is True:
            pc_headings = ppc.headings
            pc_zenits = ppc.zenits
            xs = np.cos(np.radians(pc_headings))
            ys = np.sin(np.radians(pc_headings))
            zs = np.sin(np.radians(pc_zenits))

            ax.quiver(ppc.pos_xs, ppc.pos_ys,ppc.pos_zs, xs, ys, zs,color='k')

    ax.axis('auto')

    plt.show()
def decode_depthmap(source, resize = True, size = (8192,4096), show = False):
    """
    Function to decode the depthmaps generated by the navvis processing
    source: Location of the PNG files containing the depthmap
    resize(bool): If the resulting dethmap needs to be resized to match the size of the corresponding pano, by default True
    size: size of the corresponding pano, by default 8192x4096
    show: if true the result wil be shown, by default False
    """
    depthmap = np.asarray(Image.open(source)).astype(float)
    converted_depthmap = np.empty([np.shape(depthmap)[0], np.shape(depthmap)[1]])
    r = 0
    while r < np.shape(depthmap)[0]:
        c = 0
        while c < np.shape(depthmap)[1]:
            value = depthmap[r,c]
            depth_value = value[0] / 256 * 256 + value[1] / 256 * 256 * 256 + value[2] / 256 * 256 * 256 * 256 + value[3] / 256 * 256 * 256 * 256 * 256
            converted_depthmap[r,c] = depth_value
            c = c + 1
        r = r + 1
    if resize:
        resized_depthmap = cv2.resize(converted_depthmap,size)
        if show:
            plt.imshow(resized_depthmap, cmap="plasma")
            plt.show()
        return resized_depthmap
    else:
        if show:
            plt.imshow(converted_depthmap, cmap="plasma")
            plt.show()
        return converted_depthmap




"""
ImageNode - a Python Class to govern the data and metadata of image data (JPG,PNG,XML,XMP)
"""
#IMPORT PACKAGES
from distutils import extension
import xml.etree.ElementTree as ET
import cv2
import PIL
from rdflib import Graph, URIRef
import numpy as np
import os
import open3d as o3d
import math

#IMPORT MODULES
from geomapi.nodes import Node
import geomapi.utils as ut
import geomapi.utils.geometryutils as gt


class OrthoNode(Node):
    # class attributes
    
    def __init__(self,  graph : Graph = None, 
                        graphPath:str=None,
                        subject : URIRef = None,
                        path : str=None, 
                        xmpPath: str = None,
                        xmlPath: str = None,
                        xmlIndex: int = None,
                        getResource : bool = False,
                        image : np.ndarray  = None, 
                        **kwargs): 
        """
        Creates an OrthoNode. Overloaded function.

        Args:
            0.graph (RDFlib Graph) : Graph with a single subject (if multiple subjects are present, only the first will be used to initialise the MeshNode)
            
            1.graphPath (str):  Graph file path with a single subject (if multiple subjects are present, only the first will be used to initialise the MeshNode)

            2.path (str) : path to .jpg or .png file (Note that this node will also contain the data)

            3.orthomosaic (ndarray) : OpenCV image as numpy ndarray (Note that this node will also contain the data)
                
        Returns:
            A OrthoNode with metadata (if you also want the data, call node.get_resource() method)
        """  
        self.path=path
        self.image = image
        self.xResolution = None # (Float) 
        self.yResolution = None # (Float) 
        self.resolutionUnit = None # (string)
        self.imageWidth = None # (int) number of pixels
        self.imageHeight = None  # (int) number of pixels
        super().__init__(   graph= graph,
                            graphPath= graphPath,
                            subject= subject,
                            **kwargs) 
        
        
        #self.gsd =gsd?
 



"""
SessionNode - a Python Class to govern the data and metadata of remote sensing data captured in the same epoch. \n

This node buids upon Open3D for its geometry definitions.\n
This node uses LinkedNodes (ImageNodes, PointCloudNodes, etc.) to describe resources gathered in the same epoch.
It directly inherits from Node.\n
Be sure to check the properties defined in the above classes to initialise the Node.

Goals:
- Given a path, import all the linked images, meshes, ect... into a session class \n
- Convert non-RDF metadata files (json, xml, ect..) to SessionsNodes and export them to RDF \n
- get the boundingbox of the whole session \n
- use URIRef() to reference the images, ect... \n
"""  
from xml.dom.minicompat import NodeList
from rdflib import Graph, URIRef, Literal
import rdflib
from rdflib.namespace import RDF
import geomapi
from typing import List
import open3d as o3d 
import cv2
import numpy as np
import os
from typing import Union
import uuid
import concurrent.futures

#IMPORT MODULES
from geomapi.nodes import *
import geomapi.utils as ut
import geomapi.utils.geometryutils as gt

class SessionNode(GeometryNode):

    def __init__(self,  graph : Graph = None, 
                        graphPath: str=None,
                        subject : URIRef = None,
                        linkedNodes : Node = None,
                        linkedResources = None,
                        getResource : bool = False,
                        getMetaData : bool = True,
                        **kwargs):
        """
        Creates a SessionNode & all the child nodes. Overloaded function.\n
        This Node can be initialised from one or more of the inputs below.\n
        By default, no data is imported in the Node to speed up processing.\n
        If you also want the data, call node.get_resource() or set getResource() to True.\n

        Args:
            0.graph (Graph, optional): Upon initialisation, bigger graphs can be submitted and parsed i.e.: \n
                > The RDF Graph of only the sessionNode (1 subject)
                > AN RDF Graph with only resourceNodes
            1.graphPath (str, optional): The path of the Graph of only the session. \n
            2.subject (URIRef, optional): The subject to parse the Graph with. \n
            3.resources (Any, optional): A set of resources which will be converted to LinkedNodes \n
            4.LinkedNodes (Nodes, optional): A set of Nodes (ImageNodes, MeshNodes) to contain within the session \n
            
            getResource (bool, optional= False): If True, the node will search for its physical resource on drive \n
            getMetaData (bool, optional= True): If True, the node will attempt to extract geometric metadata from the resource if present (cartesianBounds, etc.) \n
        
        Returns:
            Args(None): Create an empty Graph and Node with an unique subject guid
            
            Args(subject): create a new Graph and Node with the given subject
            Args(subject & graphPath): parse the graph with the given subject
                - 1 or more are matched: Use that node as the SessionNode
                - 1 or more are found but not matched: Raise an error
                - None are found: Create a new node with the given subject

            OVERWRITE
            Args(Graph):  parse the graph and search for a SessionNode
                - 1 or more are found: Use that node as the SessionNode
                - None are found: Create a new SessionNode with an unique id containing all the other Nodes
            Args(grapPath): parse the graph and search for a SessionNode
                - 1 or more are found: Use that node as the SessionNode
                - None are found: Create a new SessionNode with an unique id containing all the other Nodes

            Args(nodeList): create a new Graph from the joint metadata
           
        """
        #private attributes 
        self._linkedNodes=[]
        self._linkedSubjects=[]
        self._subject=None
        self._graph=None
        self._graphPath=None
        self._path=None
        self._name=None

        #instance variables
        self.linkedNodes=linkedNodes    
        self.graphPath=graphPath     
        self.subject=subject

        #initialisation functionality
        if(graphPath and not graph):
            graph = Graph().parse(graphPath)

        if(graph):
            self.parse_session_graph(graph) 

        if (self._subject is None):
            self.subject=str(uuid.uuid1())

        super().__init__(graph=self._graph, 
                        graphPath=self._graphPath, 
                        subject=self._subject,
                        **kwargs)
        
        if(linkedResources):
            self.resources_to_linked_nodes(linkedResources)

        if(getResource):
            self.get_resource()

        if (getMetaData):
            self.get_metadata_from_linked_nodes()
#---------------------PROPERTIES----------------------------

    #---------------------linkedNodes----------------------------
    @property
    def linkedNodes(self): 
        """Get the linkedNodes (Node) of the node."""
        return self._linkedNodes

    @linkedNodes.setter
    def linkedNodes(self,list):
        list=ut.item_to_list(list)
        if not list or list[0] is None:
            return []
        elif all('Node' in str(type(value)) for value in list):
            self.add_linked_nodes(list)
        else:
            raise ValueError('Some elements in self.linkedNodes are not Nodes')    
    
    #---------------------linkedSubjects----------------------------
    @property
    def linkedSubjects(self): 
        """Get the linkedSubjects (URIRef) of the node."""
        if not self._linkedSubjects:
           self._linkedSubjects= [node.subject for node in self.linkedNodes]
        return self._linkedSubjects

    @linkedSubjects.setter
    def linkedSubjects(self,list):
        list=ut.item_to_list(list)
        if not list or list[0] is None:
            return []
        elif all('URIRef' in str(type(value)) for value in list):
            matches=[w for w in self._linkedSubjects if w in list]
            (self._linkedSubjects.remove(match) for match in matches)
            self._linkedSubjects.extend(list)
        else:
            for value in list:
                if not ut.check_if_uri_exists(self._linkedSubjects,URIRef(value)):
                    string=str(value)                
                    prefix='file:///'
                    if 'file:///' in string:
                        string=string.replace('file:///','')
                        prefix='file:///'
                    elif 'http://' in string:
                        string=string.replace('http://','')
                        prefix='http://'                     
                    self._linkedSubjects.append(URIRef(prefix+ut.validate_string(string)) )
#---------------------METHODS----------------------------

    def get_linked_nodes(self,resourceGraph:Graph,linkedSubjects:List[URIRef] = None):
        """Get self.linkedNodes based on a resourceGraph and list of linkedSubjects or self.linkedSubjects.

        Args:
            resourceGraph (Graph): contains the target nodes' metadata
            linkedSubjects (List[URIRef],optional)

        Returns:
            self.linkedNodes (Node)
        """

        #validate inputs
        if(not type(resourceGraph) ==Graph): 
            raise ValueError('resourceGraph is not an RDFlib Graph')
        if(not linkedSubjects and self.linkedSubjects): 
            linkedSubjects=self.linkedSubjects
        elif(not linkedSubjects and not self.linkedSubjects):
            print('No linkedSubjects present, taking all resource subjects')
            self.linkedSubjects=[s for s in resourceGraph.subjects(RDF.type)]

        nodeSubjectList=[node.subject for node in self.linkedNodes if (self.linkedNodes)]
        #create nodes
        for subject in linkedSubjects:   
            s=subject.toPython() #temp   
            if not ut.check_if_uri_exists(nodeSubjectList,subject):                
                g=ut.get_subject_graph(resourceGraph,subject=subject)  
                if (g):            
                    newNode=create_node(graph=g, subject=subject)
                    self.linkedNodes.append(newNode)
        return self.linkedNodes
    
    def add_linked_subjects(self,linkedSubjects : List[URIRef]=None):
        """Update self.linkedSubjects

        Args:
            linkedSubjects (List[URIRef], optional): 
        """   
        linkedSubjects=ut.item_to_list(linkedSubjects)
        [self.linkedSubjects.append(s) for s in linkedSubjects if s not in self.linkedSubjects]
      
    def add_linked_nodes(self,linkedNodes:List[Node]):
        """Update self.linkedNodes from a new linkedNodes list

        Args:
            linkedNodes (List[Node]): 
        """        
        linkedNodes=ut.item_to_list(linkedNodes) 
        for node in linkedNodes:
            if node not in self.linkedNodes:
                self.linkedNodes.append(node)
        self.add_linked_subjects([node.subject for node in linkedNodes])

    def parse_session_graph(self,graph:Graph)-> Union[URIRef , Graph, Node]:
        """Parse a graph to detect and split sessionNodes from resourceNodes

        Args:
            graph (Graph): Graph that either contains a sessionNode, resourceNodes or a combination of both. 

        Raises:
            ValueError: If self.subject is provided, it should match with a subject in the graph

        Returns:
            Union[subject ,Graph, Nodelist]
        """        
        # Step 1: extract the SessionNodes
        subjects = graph.subjects(RDF.type)
        sessionNodeSubjects = []
        resourceNodeSubjects = []
        for sub in subjects:
            nodeType = ut.literal_to_string(graph.value(subject=sub,predicate=RDF.type))
            if 'SessionNode' in nodeType:
                sessionNodeSubjects.append(sub)
            else:
                resourceNodeSubjects.append(sub)
        
        # Step 2: Get the SessionNode in the graph (if it exists)
        if (not sessionNodeSubjects): 
            print("no sessionSubjects found")
            self.get_subject()

        else: # there is 1 or more sessionNodes, search for a match
            if (not self._subject): # no subject was given, pick one from the list
                self.subject = sessionNodeSubjects[0]
                self.graph=ut.get_subject_graph(graph=graph,subject=self._subject)               
                if(len(sessionNodeSubjects) > 1): 
                    print("More than one SessionNode is present, while no subject was provided, picked:",self.subject,"out of", sessionNodeSubjects)
            else: # Check if the subject is in the list
                if (self._subject not in sessionNodeSubjects):
                    raise ValueError("The given subject is not in the Graph or is not a SessionNode")

        # Step 3: Parse all the other Nodes into the nodelist
        nodelist=[]
        for subject in resourceNodeSubjects:
            s=subject.toPython() #temp
            newGraph=ut.get_subject_graph(graph=graph,subject=subject)
            nodelist.append(create_node(graph = newGraph ,graphPath= self.graphPath,subject= subject))
        if nodelist:
            self.add_linked_nodes(nodelist)
        
        return self.subject, self.graph, self.linkedNodes
    
    def linked_nodes_to_graph(self,graphPath:str=None,save:bool=False) -> Graph:
        """Serialize the session's linkedNodes

        Args:
            graphPath (str,optional): defaults to linkednodesGraph.ttl in session graphPath location.
            save (bool, optional): Defaults to False.

        Returns:
            Graph with linkedNodes
        """
        if not graphPath:
            if self.graphPath:
                graphPath=os.path.join(ut.get_folder(self.graphPath),'linkednodesGraph.ttl')
            elif self.get_path() and os.path.exists(self.path):
                graphPath=os.path.join(self.path,'linkednodesGraph.ttl')

        g=Graph()
        g=ut.bind_ontologies(g)
        for node in self.linkedNodes:
                node.to_graph(graphPath)
                g+= node.graph
        if(graphPath and save):
            g.serialize(graphPath)     
        return g  

    def session_to_graph(self,graphPath:str=None,save:bool=False) -> Graph:
        """Serialize the session's linkedNodes and own metadata

        Args:
            graphPath (str,optional): defaults to combinedGraph.ttl in session graphPath location.
            save (bool, optional): Defaults to False.

        Returns:
            Graph with linkedNodes and sessionNode
        """
        if graphPath and next(graphPath.endswith(extension) for extension in ut.RDF_EXTENSIONS) :
            self._graphPath=graphPath

        if not graphPath:
            if self.graphPath:
                graphPath=os.path.join(ut.get_folder(self.graphPath),'combinedGraph.ttl')
            elif self.get_path() and os.path.exists(self.path):
                graphPath=os.path.join(self.path,'combinedGraph.ttl')

        g=Graph()
        g=ut.bind_ontologies(g)
        for node in self.linkedNodes:
            node.to_graph(graphPath)
            g+= node.graph
        g+=self.to_graph(graphPath)
        if(graphPath and save):
            g.serialize(graphPath)     
        return g 

    def save_linked_resources(self,directory:str=None):
        """Export the resources of the linkedNodes.\n

        Args:
            directory (str, optional): directory folder to store the data.\n

        Returns:
            bool: return True if export was succesful
        """ 
        for node in self.linkedNodes:
            node.save_resource(directory)  

    def save_resource(self,directory:str=None,extension :str = '.ply') -> bool:
        """Export the resource (Convex hull) of the Node.\n

        Args:
            directory (str, optional): directory folder to store the data.\n
            extension (str, optional): file extension. Defaults to '.ply'.\n

        Raises:
            ValueError: Unsuitable extension. Please check permitted extension types in utils._init_.\n

        Returns:
            bool: return True if export was succesful
        """ 
        #check path
        if self.resource is None:
            return False
        
        #validate extension
        if extension not in ut.MESH_EXTENSION:
            raise ValueError('Invalid extension')

        # check if already exists
        if directory and os.path.exists(os.path.join(directory,self.get_name() + extension)):
            self.path=os.path.join(directory,self.get_name() + extension)
            return True
        elif not directory and self.get_path() and os.path.exists(self.path) and extension in ut.MESH_EXTENSION:            
            return True
                    
        #get directory
        if (directory):
            pass    
        elif self.path is not None:    
            directory=ut.get_folder(self.path)            
        elif(self.graphPath): 
            dir=ut.get_folder(self.graphPath)
            directory=os.path.join(dir,'SESSION')   
        else:
            directory=os.path.join(os.getcwd(),'SESSION')
        # create directory if not present
        if not os.path.exists(directory):                        
            os.mkdir(directory) 

        self.path=os.path.join(directory,ut.get_filename(self.subject.toPython()) + extension)

        #write files
        if o3d.io.write_triangle_mesh(self.path, self.resource):
            return True
        return False

    def resources_to_linked_nodes(self,linkedResources) -> Node:
        """Create proper node type from a data resource

        Args:
            resource (mesh,pcd,image,ortho,lineset,etc.): 

        Returns:
            Node: 
        """
        nodelist=[]
        for resource in linkedResources:
            #check type
            if 'TriangleMesh' in str(type(resource)):
                nodelist.append(MeshNode(resource=resource))
            elif type(resource) is np.ndarray:
                nodelist.append( ImageNode(resource=resource))
            elif 'PointCloud' in str(type(resource)):
                nodelist.append( PointCloudNode(resource=resource))
            elif 'LineSet' in str(type(resource)):
                nodelist.append( LinesetNode(resource=resource))
            elif 'ifcelement' in str(type(resource)):
                nodelist.append( BIMNode(resource=resource))
        if nodelist:
            self.add_linked_nodes(nodelist)

    def get_metadata_from_linked_nodes(self):
        """Returns the sessionNode metadata from the linkedNodes. \n

        Features:
            cartesianTransform\n
            orientedBoundingBox \n
            orientedBounds\n
            cartesianBounds\n
            resource \n

        Returns:
            bool: True if exif data is successfully parsed
        """
        if (getattr(self,'cartesianTransform',None) is not None and
            getattr(self,'orientedBoundingBox',None) is not None and
            getattr(self,'orientedBounds',None) is not None and
            getattr(self,'cartesianBounds',None) is not None and
            getattr(self,'resource',None) is not None):
            return True

        if getattr(self,'timestamp',None) is None and self.linkedNodes:
            self.timestamp=self.linkedNodes[0].get_timestamp()

        points=o3d.cpu.pybind.utility.Vector3dVector()        
        for node in self.linkedNodes: 
            if getattr(node,'get_oriented_bounding_box',None) is not None:           
                box=node.get_oriented_bounding_box()
                if (box):
                    points.extend(box.get_box_points())    
                
            elif getattr(node,'cartesianTransform',None) is not None:
                t=gt.get_translation(node.get_cartesian_transform())
                t=np.reshape(t,(1,3))
                p=o3d.cpu.pybind.utility.Vector3dVector(t)
                points.extend(p)

        if np.asarray(points).shape[0] >=5:    
            self.cartesianTransform=gt.get_cartesian_transform(translation=np.mean( np.asarray(points),0)) 
            self.orientedBoundingBox=o3d.geometry.OrientedBoundingBox.create_from_points(points)
            self.orientedBounds=np.asarray(self.orientedBoundingBox.get_box_points())
            self.cartesianBounds=gt.get_cartesian_bounds(o3d.geometry.AxisAlignedBoundingBox.create_from_points(points))
            pcd= o3d.geometry.PointCloud()
            pcd.points=points
            hull, _ =pcd.compute_convex_hull()
            self.resource=hull

    def get_linked_resources(self,percentage:float=1.0):
        """Returns the resources of the linkedNodes. \n
        If none is present, it will search for the data on drive from path, graphPath, name or subject. \n
        Otherwise, it will be reconstructed from the metadata present

        Args:
            1. self (sessionNode)
            2. percentage(float,optional): load percentage of point cloud resources in present PointCloudNodes.

        Returns:
            list[resource] or None
        """
        for node in self.linkedNodes:
            if 'PointCloud' in str(type(node)):
                node.get_resource(percentage)
            else:
                node.get_resource()
        return [n.resource for n in self.linkedNodes]
      
    def get_linked_resources_multiprocessing(self,percentage:float=1.0):
        """Returns the resources of the linkedNodes by multi-processing the imports. \n
        If none is present, it will search for the data on drive from path, graphPath, name or subject. \n

        **NOTE**: Starting parallel processing takes a bit of time. As such, this method will only outperform get_linked_resources with 10+ linkedNodes

        Args:
            1. self (sessionNode)
            2. percentage(float,optional): load percentage of point cloud resources in present PointCloudNodes.

        Returns:
            list[resource] or None
        """
        [n.get_path() for n in self.linkedNodes]

        with concurrent.futures.ProcessPoolExecutor() as executor:
            # first load all data and output it as np.arrays      
            results1=[executor.submit(gt.e57_to_arrays,e57Path=n.path,e57Index=n.e57Index,percentage=percentage,tasknr=i) for i,n in enumerate(self.linkedNodes) 
                        if 'PointCloud' in str(type(n)) and 
                            n.resource is None and
                            n.path.endswith('.e57') and 
                            os.path.exists(n.path)]

            results2=[executor.submit(gt.pcd_to_arrays,path=n.path,percentage=percentage,tasknr=i) for i,n in enumerate(self.linkedNodes) 
                        if 'PointCloud' in str(type(n)) and 
                            n.resource is None and
                            n.path.endswith('.pcd') and 
                            os.path.exists(n.path)]
            results3=[executor.submit(gt.mesh_to_arrays,path=n.path,tasknr=i) for i,n in enumerate(self.linkedNodes) 
                        if ('MeshNode' in str(type(n)) or 'BIMNode' in str(type(n))) and 
                            n.resource is None and
                            os.path.exists(n.path)]
            results4=[executor.submit(gt.img_to_arrays,path=n.path,tasknr=i) for i,n in enumerate(self.linkedNodes) 
                        if 'ImageNode' in str(type(n)) and 
                        n.resource is None and
                        os.path.exists(n.path)]

        # next, the arrays are assigned to point clouds outside the loop.
        # Note that these loops should be in parallel as they would otherwise obstruct one another.
        for r1 in concurrent.futures.as_completed(results1): 
            resource=gt.arrays_to_pcd(r1.result())
            self.linkedNodes[r1.result()[-1]].resource=resource
        for r2 in concurrent.futures.as_completed(results2): 
            resource=gt.arrays_to_pcd(r2.result())
            self.linkedNodes[r2.result()[-1]].resource=resource
        for r3 in concurrent.futures.as_completed(results3): 
            # print(len(r3.result()[0]))
            resource=gt.arrays_to_mesh(r3.result())
            self.linkedNodes[r3.result()[-1]].resource=resource
        for r4 in concurrent.futures.as_completed(results4): 
            self.linkedNodes[r4.result()[-1]].resource=r4.result()[0]
        return [n.resource for n in self.linkedNodes]

    def get_resource(self)->o3d.geometry.TriangleMesh: 
        """Returns the convexhull of the node. \n
        If none is present, it will search for the data on drive from path, graphPath, name or subject. \n

        Returns:
            o3d.geometry.TriangleMesh or None
        """
        if self.resource is not None and len(self.resource.triangles)>=2:
            pass
        elif self.get_path():
            self._resource   =  o3d.io.read_triangle_mesh(self.path)
        elif self.linkedNodes:
            points=o3d.cpu.pybind.utility.Vector3dVector()
            for node in self.linkedNodes: 
                if getattr(node,'get_bounding_box',None):           
                    box=node.get_bounding_box()
                    if (box):
                        points.extend(box.get_box_points()) 
            pcd= o3d.geometry.PointCloud()
            pcd.points=points
            self.resource=pcd.compute_convex_hull()   
        return self.resource  

    
    def get_metadata_from_resource(self) -> bool:
        """Returns the metadata from a resource. \n

        Features:
            cartesianTransform\n
            orientedBoundingBox\n
            cartesianBounds\n
            orientedBounds \n

        Returns:
            bool: True if exif data is successfully parsed
        """
        if (not self.resource or
            len(self.resource.triangles) <2):
            return False    

        try:
            if  getattr(self,'cartesianTransform',None) is None:
                center=self.resource.get_center()  
                self.cartesianTransform= np.array([[1,0,0,center[0]],
                                                    [0,1,0,center[1]],
                                                    [0,0,1,center[2]],
                                                    [0,0,0,1]])

            if getattr(self,'cartesianBounds',None) is  None:
                self.cartesianBounds=gt.get_cartesian_bounds(self.resource)
            if getattr(self,'orientedBoundingBox',None) is  None:
                self.orientedBoundingBox=self.resource.get_oriented_bounding_box()
            if getattr(self,'orientedBounds',None) is  None:
                box=self.resource.get_oriented_bounding_box()
                self.orientedBounds= np.asarray(box.get_box_points())
            return True
        except:
            raise ValueError('Metadata extraction from resource failed')
           
#################################
def create_node(graph: Graph = None, graphPath: str =None, subject: URIRef = None, resource = None, **kwargs)-> Node:
    """_summary_

    Args:
        graph (Graph, optional): _description_. Defaults to None.
        graphPath (str, optional): _description_. Defaults to None.
        subject (URIRef, optional): _description_. Defaults to None.

    Returns:
        Node (PointCloudNode,MeshNode,GeometryNode,ImageNode)
    """
    #input validation
    if(graphPath and not graph):
            graph = Graph().parse(graphPath)
    if(graph and not subject):
        subject=next(graph.subjects(RDF.type))
    if (subject and graph):    
        nodeType = ut.literal_to_string(graph.value(subject=subject,predicate=RDF.type))
    elif (resource):
        if type(resource) is o3d.geometry.PointCloud:
            nodeType='PointCloudNode'
        elif type(resource) is o3d.geometry.TriangleMesh:
            nodeType='MeshNode'
        elif type(resource) is o3d.geometry:
            nodeType='GeometryNode'
        elif type(resource) is np.ndarray:
            nodeType='ImageNode'        
    else:        
        nodeType = 'Node'

    #node creation
    if 'BIMNode' in nodeType:
        node=BIMNode(graph=graph, graphPath=graphPath, resource=resource,subject=subject, **kwargs)
    elif 'MeshNode' in nodeType:
        node=MeshNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)
    elif 'GeometryNode' in nodeType:
        node=GeometryNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)
    elif 'PointCloudNode' in nodeType:
        node=PointCloudNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)
    elif 'ImageNode' in nodeType:
        node=ImageNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)
    elif 'SessionNode' in nodeType:
        node=SessionNode(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs)  
    else:
        node=Node(graph=graph, graphPath=graphPath, resource=resource, subject=subject, **kwargs) 
    return node



"""
BIMNode is a Python Class to govern the data and metadata of BIM data. 
This node builds upon the Open3D and IFCOpenShell API for the BIM definitions.
It inherits from GeometryNode which in turn inherits from Node.
Be sure to check the properties defined in those abstract classes to initialise the Node.

**IMPORTANT**: The current BIMNode class is designed from a geospatial perspective to 
use in geometric analyses. As such, it's geometry is defined by Open3D.geometry.TriangleMesh objects 
and contains only a skeleton set of IFC Information. Users should use this class to conduct their analyses
and then combine it with the existing IFC files or IFCOWL RDF variants to integrate the results.

"""
#IMPORT PACKAGES
# from ast import Raise
import open3d as o3d 
import numpy as np 
from rdflib import Graph, URIRef
import os
import ifcopenshell
import ifcopenshell.geom as geom
import ifcopenshell.util
from ifcopenshell.util.selector import Selector


#IMPORT MODULES
from geomapi.nodes import GeometryNode
import geomapi.utils as ut
import geomapi.utils.geometryutils as gt

class BIMNode (GeometryNode):
    def __init__(self,  graph : Graph = None, 
                        graphPath: str= None,
                        subject : URIRef = None,
                        path : str= None, 
                        ifcPath : str = None,                        
                        globalId : str = None,
                        getResource : bool = False,
                        getMetaData : bool = True,
                        **kwargs): 
        """Creates a BIMNode from one or more of the following inputs. 
        By default, no data is imported in the Node to speed up processing.
        If you also want the data, call node.get_resource() or set getResource() to True.\n

        Args:\n 
            0.subject (RDFlib URIRef): subject to be used as the main identifier in the RDF Graph\n
            1.graph (RDFlib Graph) : RDF Graph with a single subject. if no subject is present, the first subject in the Graph will be used to initialise the Node.\n     
            2.graphPath (str):  RDF Graph file path. if no subject is present, the first subject in the Graph will be used to initialise the Node.\n
            3.path (str) : path to mesh file (Note that this node will also contain the data)\n
            4.resource (o3d.geometry.TriangleMesh, ifcopenshell.entity_instance): Warning, never attach an IfcElement to a node directly as this is very unstable! \n
            5.ifcPath (str) : path to IFC file\n
            6.globalId (str) : IFC globalId\n
            \n
            - getResource (bool, optional= False): If True, the node will search for its physical resource on drive \n
            - getMetaData (bool, optional= True): If True, the node will attempt to extract geometric metadata from the resource if present (cartesianBounds, etc.) \n
                
        Returns:
            A BIMNode with metadata 
        """           
        #private attributes 
        self._ifcPath=None
        self._globalId=None

        super().__init__(   graph= graph,
                            graphPath= graphPath,
                            subject= subject,
                            path=path,
                            **kwargs) 
                            
        #instance variables
        self.ifcPath=ifcPath
        self.globalId=globalId

        #initialisation functionality
        if getResource:
            self.get_resource() 
        
        if getMetaData:
            self.get_metadata_from_ifc_path()
            self.get_metadata_from_resource()

#---------------------PROPERTIES----------------------------

    #---------------------ifcPath----------------------------
    @property
    def ifcPath(self): 
        """Get the ifcPath (str) of the node."""
        return self._ifcPath

    @ifcPath.setter
    def ifcPath(self,value):
        if value is None:
            return None
        if (type(value) is str and ut.get_extension(value) =='.ifc'):
            self._ifcPath=value
        else:
            raise ValueError('self.ifcPath has invalid type, path or extension')

    #---------------------globalId----------------------------
    @property
    def globalId(self): 
        """Get the globalId (str) of the node."""
        return self._globalId

    @globalId.setter
    def globalId(self,value):
        if value is None:
            return None
        try: 
            self._globalId=str(value)
        except:
            raise TypeError('self.globalId should be string compatible')

#---------------------METHODS----------------------------
    def set_resource(self,value):
        """Set self.resource (o3d.geometry.TriangleMesh) of the Node.\n

        Args:
            1. o3d.geometry.TriangleMesh \n
            2. trimesh.base.Trimesh\n
            3. ifcopenshell.entity_instance (this also sets the name, subject, etc.\n

        Raises:
            ValueError: Resource must be ao3d.geometry.TriangleMesh, trimesh.base.Trimesh or ifcopenshell.entity_instance with len(mesh.triangles) >=2.
        """
        if 'TriangleMesh' in str(type(value)) and len(value.triangles) >=2:
            self._resource = value
        elif 'Trimesh' in str(type(value)):
            self._resource=  value.as_open3d
        elif type(value) is ifcopenshell.entity_instance:
            self._resource= gt.ifc_to_mesh(value)
            self.name=value.Name
            self.className=value.is_a()
            self.globalId=value.GlobalId
            if self.name and self.globalId:
                self.subject= self.name +'_'+self.globalId 
        else:
            raise ValueError('Resource must be ao3d.geometry.TriangleMesh, trimesh.base.Trimesh or ifcopenshell.entity_instance with len(mesh.triangles) >=2')

    def get_resource(self)->o3d.geometry.TriangleMesh: 
        """Returns the mesh data in the node.
        If none is present, it will search for the data on using the attributes below.

        Features:
            1. self.path\n
            2. self.graphPath\n
            3. self.name or self.subject\n

        Returns:
            o3d.geometry.TriangleMesh or None
        """
        if self.resource is not None and len(self.resource.triangles)>=2:
            pass
        elif self.get_path():
            resource =  o3d.io.read_triangle_mesh(self.path)
            if len(resource.triangles)>2:
                self._resource  =resource
        elif self.ifcPath and os.path.exists(self.ifcPath):
            try:
                ifc = ifcopenshell.open(self.ifcPath)   
                ifcElement= ifc.by_guid(self.get_globalId())
                self._resource=gt.ifc_to_mesh(ifcElement)
            except:
                print('mesh=gt.ifc_to_mesh(ifcElement) error')
        return self.resource  

    def get_globalId(self):
        """Returns the ifc globalId (str).
        """
        if self._globalId:
            pass 
        elif os.path.exists(self.ifcPath):
            selector = Selector()
            ifc = ifcopenshell.open(self.ifcPath)  
            ifcElement=next(ifcElement for ifcElement in selector.parse(ifc, '.ifcObject') )
            self._globalId=ifcElement.GlobalId
        return self._globalId

    def save_resource(self, directory:str=None,extension :str = '.ply') ->bool:
        """Export the resource of the Node.\n

        Args:
            1. directory (str, optional): directory folder to store the data.\n
            2. extension (str, optional): file extension. Defaults to '.ply'.\n

        Raises:
            ValueError: Unsuitable extension. Please check permitted extension types in utils._init_.\n

        Returns:
            bool: return True if export was succesful
        """          
        #check path
        if self.resource is None:
            return False
        
        #validate extension
        if extension not in ut.MESH_EXTENSION:
            raise ValueError('Invalid extension')

        # check if already exists
        if directory and os.path.exists(os.path.join(directory,self.get_name() + extension)):
            self.path=os.path.join(directory,self.get_name() + extension)
            return True
        elif not directory and self.get_path() and os.path.exists(self.path) and extension in ut.MESH_EXTENSION:
            return True
                    
        #get directory
        if (directory):
            pass    
        elif self.path is not None:    
            directory=ut.get_folder(self.path)            
        elif(self.graphPath): 
            dir=ut.get_folder(self.graphPath)
            directory=os.path.join(dir,'BIM')   
        else:
            directory=os.path.join(os.getcwd(),'BIM')
        # create directory if not present
        if not os.path.exists(directory):                        
            os.mkdir(directory) 

        self.path=os.path.join(directory,ut.get_filename(self.get_name()) + extension) #subject.toPython() replaced by get_name()

        #write files
        if o3d.io.write_triangle_mesh(self.path, self.resource):
            return True
        return False
    
    def get_metadata_from_ifc_path(self) -> bool:
        """Returns the metadata from a resource. \n

        Args (self):
            ifcPath\n
            globalId\n

        Features:
            1. PointCount\n
            2. faceCount \n
            3. cartesianTransform\n
            4. cartesianBounds\n
            5. orientedBounds \n
            6. globalId \n
            7. name \n

        Returns:
            bool: True if exif data is successfully parsed
        """        
        if (not self.ifcPath or 
            not os.path.exists(self.ifcPath) or
            not self.get_globalId()):
            return False
        
        if (getattr(self,'name',None) is not None and
            getattr(self,'className',None) is not None):
            return True
        
        ifc = ifcopenshell.open(self.ifcPath)   
        ifcElement= ifc.by_guid(self.globalId)
        if ifcElement:
            self.name=ifcElement.Name 
            self.className=ifcElement.is_a()   
            if self.name and self.globalId:
                self.subject= self.name +'_'+self.globalId 
            return True
        else:
            return False
     
    def get_metadata_from_resource(self) -> bool:
        """Returns the metadata from a resource. \n

        Features:
            1. PointCount\n
            2. faceCount \n
            3. cartesianTransform\n
            4. cartesianBounds\n
            5. orientedBounds \n

        Returns:
            bool: True if exif data is successfully parsed
        """
        if (not self.resource or
            len(self.resource.triangles) <2):
            return False    

        try:
            if getattr(self,'pointCount',None) is None:
                self.pointCount=len(self.resource.vertices)

            if getattr(self,'faceCount',None) is None:
                self.faceCount=len(self.resource.triangles)

            if  getattr(self,'cartesianTransform',None) is None:
                center=self.resource.get_center()  
                self.cartesianTransform= np.array([[1,0,0,center[0]],
                                                    [0,1,0,center[1]],
                                                    [0,0,1,center[2]],
                                                    [0,0,0,1]])

            if getattr(self,'cartesianBounds',None) is  None:
                self.cartesianBounds=gt.get_cartesian_bounds(self.resource)
            if getattr(self,'orientedBoundingBox',None) is  None:
                self.orientedBoundingBox=self.resource.get_oriented_bounding_box()
            if getattr(self,'orientedBounds',None) is  None:
                box=self.resource.get_oriented_bounding_box()
                self.orientedBounds= np.asarray(box.get_box_points())
            return True
        except:
            raise ValueError('Metadata extraction from resource failed')
       


"""
**Node** is an abstract Python Class to govern the data and metadata of remote sensing data (pcd, images, meshes, orthomozaik).
It is the base class for all other node classes. It contains the base RDF graph functionality and i/o from and to RDF files. \n
   
**IMPORTANT**: the Node class is an archytype class from which specific data classes (e.g. PointCloudNode) inherit.
Do not use this class directly if you can use a child class with more functionality.

Goals:\n
- Govern the metadata and geospatial information of big data files in an accesible and lightweight format. \n
- Serialize the metadata of remote sensing and geospatial files (BIM, Point clouds, meshes, etc.) as RDF Graphs  \n
- Attach spatial en temporal analyses through RDF Graph navigation   \n
    
"""

#IMPORT PACKAGES
from operator import contains
import os
from posixpath import relpath
import re
from typing import List
import uuid
import datetime
import numpy as np

from rdflib import Graph, URIRef, Literal
import rdflib
from rdflib.namespace import RDF

#IMPORT MODULES
import geomapi.utils as ut

class Node:
    def __init__(self,  subject: URIRef = None,
                        graph: Graph = None,
                        graphPath: str = None,                        
                        name:str=None,
                        path:str=None,
                        timestamp:str=None,
                        resource=None,
                        cartesianTransform:np.ndarray=None,
                        **kwargs):
        """Creates a Node from one or more of the following inputs. \n

        Args:
            1.graph (Graph, optional): An RDF Graph to parse.\n
            2.graphPath (str, optional): The path of an RDF Graph to parse. If no subject is provided, the first subject of the graph is retained. \n
            3.path(str,optional): A filepath to a resource. \n
            4.subject (URIRef, optional): A subject to use as identifier for the Node. If a graph is also present, the subject should be part of the graph.\n
            5.name (str, optional): A name of the Node. This is not a unique identifier but serves as non-functional description.\n

        Returns:
            Node
        """
        #private attributes 
        self._subject=None
        self._graph=None
        self._graphPath=None 
        self._path=None
        self._name=None
        self._timestamp=None 
        self._resource=None 
        self._cartesianTransform=None

        #instance variables
        self.subject=subject
        self.graphPath=graphPath
        self.graph=graph
        self.path=path        
        self.name=name
        self.timestamp=timestamp
        self.resource=resource 
        self.cartesianTransform=cartesianTransform

        #initialisation functionality
        if self._path:
            self.name=ut.get_filename(path)
            if  os.path.exists(self._path):
                self.timestamp=ut.get_timestamp(path) 

        if(graphPath and os.path.exists(self._graphPath) and not self._graph):
            self._graph = Graph().parse(graphPath) 

        self.get_subject()      
        if(self._graph):
            if ut.check_if_subject_is_in_graph(self._graph,self._subject):
                self._graph=ut.get_subject_graph(self._graph,self._subject)
                self.get_metadata_from_graph(self._graph,self._subject) 
            elif 'session' in str(type(self)):
                pass
            else:
                raise ValueError( 'Subject not in graph')
        self.__dict__.update(kwargs)

#---------------------PROPERTIES----------------------------

    #---------------------PATH----------------------------
    @property
    def path(self): 
        """Get the resource path (str) of the node. If no path is present, you can use get_path() to reconstruct the path from either 
        the graphPath or working directory
        
        Features:
            1. folder\n
            2. self.name\n
            3. self.graphPath\n
        """        
        return self._path
    
    @path.setter
    def path(self,path):
        if path is None:
            return None
        nodeExtensions=ut.get_node_resource_extensions(str(type(self)))
        if (type(path) is str and ut.get_extension(path) in nodeExtensions):
            self._path=path
        else:
            raise ValueError('self.path has invalid type, path or extension')

    #---------------------NAME----------------------------
    @property
    def name(self):
        """Get the name (str) of the node. This can include characters that the operating
        system does not allow. If no name is present, you can use get_name() to construct a name from the subject or path.

        Features:
            1. self.path\n
            2. self.subject\n
        """        
        return self._name

    @name.setter
    def name(self,name):
        if name is None:
            return None
        try: 
            self._name=str(name)
        except:
            raise TypeError('self.name should be string compatible')

    #---------------------TIMESTAMP----------------------------
    @property
    def timestamp(self):
        """Get the timestamp (str(yyyy-MM-ddTHH:mm:ss)) of the node. If no timestamp is present, use get_timestamp() to gather the timestamp from the path or graphPath.

        Features:
            1. self.path\n
            2. self.graphPath\n
        """
        return self._timestamp

    @timestamp.setter
    def timestamp(self,timestamp):
        if timestamp is None:
            return None
        elif timestamp:
            self._timestamp=ut.validate_timestamp(timestamp)
        else:
            raise ValueError('timestamp should be str(yyyy-MM-ddTHH:mm:ss)')

    #---------------------GRAPHPATH----------------------------    
    @property
    def graphPath(self):
        """Get the path (str) of graph of the node, or the graphPath in which the subject is contained."""
        return self._graphPath

    @graphPath.setter
    def graphPath(self,graphPath):
        if graphPath is None:
            return None
        elif (type(graphPath) == str and 
              next(graphPath.endswith(extension) for extension in ut.RDF_EXTENSIONS) ):
            self._graphPath=graphPath
        else:
            raise ValueError('self.graphPath has invalid type, path or extension')    

    #---------------------GRAPH----------------------------    
    @property
    def graph(self):
        """Get the graph (RDFLib.Graph) of the node. If no graph is present, you can use get_graph() to parse the graph from a graphPath. Alternatively,
        you can use to_graph() to serialize the Nodes attributes to RDF.
        
        Features:
            1. self.graphPath
        """       
        return self._graph

    @graph.setter
    def graph(self,graph):
        if graph is None:
            return None
        elif (type(graph) is rdflib.Graph):
            self._graph=graph
        else:
            raise TypeError('type(graph) should be rdflib.Graph')    

    #---------------------SUBJECT----------------------------    
    @property
    def subject(self):
        """Get the subject (RDFLib.URIRef) of the node. If no subject is present, you can use get_subject() to construct it from a graph, name or path.
        Otherwise, a random guid is generated.
        
        Features:
            1. self.name\n
            2. self.graph\n
            3. self.path\n
        """
        return self._subject

    @subject.setter
    def subject(self,subject):
        if subject is None:
            return None
        elif type(subject) is rdflib.URIRef:
            self._subject=subject
        else:
            string=str(subject)
            prefix='file:///'
            if 'file:///' in string:
                string=string.replace('file:///','')
                prefix='file:///'
            elif 'http://' in string:
                string=string.replace('http://','')
                prefix='http://' 
            self._subject=URIRef(prefix+ut.validate_string(string))  
    
    #---------------------RESOURCE----------------------------    
    @property
    def resource(self):
        """Get the resource (mesh, pcd, etc.) of the node. If no resource is present, you can use get_resource() to load the resource from a path or search it through the name and graphpath. 

        Features:
            1. self.path\n
            2. self.name\n
            3. self.graphPath\n
        """        
        return self._resource

    @resource.setter
    def resource(self,value):
        if value is None:
            return None
        else:
            self.set_resource(value)

    @resource.deleter
    def resource(self):
        self._resource=None
        #This will be depricated. use resource instead
        if getattr(self,'mesh',None) is not None:
            self._mesh=None
        if getattr(self,'image',None) is not None:
            self._image=None
        if getattr(self,'pcd',None) is not None:
            self._pcd=None
        if getattr(self,'ortho',None) is not None:
            self._ortho=None
         
    #---------------------CARTESIANTRANSFORM----------------------------    
    @property
    def cartesianTransform(self):
        """Get the Cartesian Transform (translation & rotation matrix) (np.ndarray(4x4)) of the node.
        Note that the initialisation from different inputs may result in different cartesianTransform values.\n
        E.g. the pose of a mesh is retrieved from the mean of the vertices,
             while the same pose initiated from the cartesianBounds equals the mean of that cartesianBounds array.
        \n
        If no cartesianTransform is present, you can use get_cartesianTransform() to construct the cartesiantransform from the resource, cartesianBounds, orientedBounds or orientedBoundingBox.     
        
        Features:
            1. self.cartesianBounds\n
            2. self.resource\n
            3. self.orientedBounds\n
            4. self.orientedBoundingBox\n
        """      
        return self._cartesianTransform

    @cartesianTransform.setter
    def cartesianTransform(self,value):
        if value is None:
            return None
        else:
            self.set_cartesianTransform(value)

#---------------------METHODS----------------------------     
    def get_metadata_from_graph(self, graph:Graph,subject:URIRef):
        """Convert the data contained in a graph to a set of node attributes.
        If the graph contains multiple subjects, it is reduced to the subject's triples. \n
        
        **NOTE**: The use of a SessionNode is advised when dealing with multi-subject graphs.\n

        Args:
            1. self.graph (RDFlib.Graph):  Graph to parse\n
            2. self.subject (RDFlib.URIRef): The subject to parse the graph for
        
        """
        if len([x for x in self._graph.subjects(RDF.type)])>1:
            self._graph=ut.get_subject_graph(graph,subject)

        for predicate, object in graph.predicate_objects(subject=subject):
            attr= ut.get_attribute_from_predicate(graph, predicate) 
            value=object.toPython()
            
            #GEOMETRY
            if attr == 'cartesianBounds':
                self.cartesianBounds=ut.literal_to_array(object) 
            elif attr == 'orientedBounds':
                self.orientedBounds=ut.literal_to_orientedBounds(object) 
            elif attr == 'cartesianTransform':
                self.cartesianTransform=ut.literal_to_cartesianTransform(object) 
            elif attr == 'geospatialTransform':
                self.geospatialTransform=ut.literal_to_array(object)    
            #PATHS
            elif re.search('path', attr, re.IGNORECASE):
                path=ut.literal_to_string(object)
                if path and self._graphPath:
                    if '..' in path:
                        path=path.strip('..\\')
                        folder=ut.get_folder_path(ut.get_folder_path(self._graphPath))
                    else:
                        folder=ut.get_folder_path(self._graphPath)
                    path=os.path.join(folder,path)
                setattr(self,attr,path)
                    
            #INT    
            elif attr in ut.INT_ATTRIBUTES:
                setattr(self,attr,ut.literal_to_int(object)) 
            #FLOAT
            elif attr in ut.FLOAT_ATTRIBUTES:
                setattr(self,attr,ut.literal_to_float(object)) 
            #LISTS
            elif attr in ut.LIST_ATTRIBUTES:
                setattr(self,attr,ut.literal_to_list(object)) 
            #LINKEDSUBEJCTS
            elif attr == 'linkedSubjects':
                # test=ut.literal_to_linked_subjects(object)
                # self.linkedSubjects=test
                setattr(self,attr,ut.literal_to_linked_subjects(object)) 
            
            #STRINGS
            else:
                setattr(self,attr,object.toPython()) 

    def get_subject(self) -> str:
        """Returns and validates the current subject. If empty, a new subject is created based on an unique GUID.\n

        Returns:
            subject (URIREF)
        """
        #subject
        if self._subject:
            pass
        # self.graph
        elif self._graph:
            self._subject=next(self._graph.subjects(RDF.type))
        #self.path
        elif self._path:
            self._name=ut.get_filename(self._path)
            self._subject=URIRef('file:///'+ut.validate_string(self._name))
        elif self._name:
            self._subject=URIRef('file:///'+ut.validate_string(self._name))
        #guid
        else:
            self._name=str(uuid.uuid1())
            self._subject=URIRef('file:///'+self._name)            
        return self._subject

    def get_timestamp(self):
        """Get the timestamp (str) of the Node. 
        This can be retrieved from user input, the path, the graphPath or the current time.\n

        Returns:
            timestamp (str): '%Y-%m-%dT%H:%M:%S'
        """
        if self._timestamp is None:
            if self._path and os.path.exists(self._path):
                self._timestamp=ut.get_timestamp(self._path)  
            elif self._graphPath and os.path.exists(self._graphPath):
                self._timestamp=ut.get_timestamp(self._graphPath)  
            else:
                self._timestamp=datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
        return self._timestamp

    def get_name(self) -> str:
        """Returns the name (str) of the Node.\n

        Returns:
           name (str)
        """
        if self._name is None:
            if self._path:
                self._name=ut.get_filename(self._path)
            else:                     
                self._name=ut.get_subject_name(self.subject)
        return self.name

    def set_cartesianTransform(self,value):
        """sets the cartesianTransform for the Node type. Overwrite this function for each node type.
        """
        # print("This is the base Node functionality, overwite for each childNode to retrieve the relevant cartesianTransform")

    def get_cartesianTransform(self):
        """Returns the cartesianTransform from the Node type. Overwrite this function for each node type."""
        # print("This is the base Node functionality, overwite for each childNode to retrieve the relevant cartesianTransform")

    def get_resource(self):
        """Returns the resource from the Node type. Overwrite this function for each node type.
        """
        # print("This is the base Node functionality, overwite for each childNode to import the relevant resource type")

    def set_resource(self,value):
        """sets the resource for the Node type. Overwrite this function for each node type.
        """
        # print("This is the base Node functionality, overwite for each childNode to import the relevant resource type")


    def clear_resource(self):
        """Clear all resources (images, pcd, meshes, ortho's, etc.) in the Node.
        """
        if getattr(self,'resource',None) is not None:
            self.resource=None

        # #TEMP
        # if getattr(self,'mesh',None) is not None:
        #     self.mesh=None
        # if getattr(self,'image',None) is not None:
        #     self.image=None
        # if getattr(self,'pcd',None) is not None:
        #     self.pcd=None
        # if getattr(self,'ortho',None) is not None:
        #     self.ortho=None
            
    def get_path(self) -> str:
        """Returns the full path of the resource from this Node.\n

        Features:
            1. self.graphPath\n
            2. self._name or self._subject\n

        Returns:
            path (str)
        """      
        if self._path and os.path.exists(self._path):
            return self._path
        
        elif self._graphPath and (self._name or self._subject):
            folder=ut.get_folder_path(self._graphPath)
            nodeExtensions=ut.get_node_resource_extensions(str(type(self)))
            allSessionFilePaths=ut.get_list_of_files(folder) 
            for path in allSessionFilePaths:
                if ut.get_extension(path) in nodeExtensions:
                    if self.get_name() in path or self.get_subject() in path :
                        self._path = path    
                        return self._path
            if self._name:
                self._path=os.path.join(folder,self._name+nodeExtensions[0])
            else:
                self._path=os.path.join(folder,self._subject+nodeExtensions[0])
            return self._path
        else:
            # print("No file containing this object's name and extension is found in the graphPath folder")
            return None

    def get_graph(self):
        """Returns the graph of the Node.

        Features:
            1. self.graphPath\n
            2. self._subject\n

        Returns:
            graph (RDFlib.GRAPH)
        """
        if self._graph is None:
            if self._graphPath and os.path.exists(self._graphPath):
                self._graph=Graph().parse(self._graphPath)
                if self._subject and ut.check_if_subject_is_in_graph(self._graph,self._subject):
                    self._graph=ut.get_subject_graph(self._graph,self._subject)
                else:
                    print( 'Subject not in Graph')
        return self._graph

    def to_graph(self, graphPath : str = None, overwrite:bool=True,save:bool=False) -> Graph:
        """Converts the current Node variables to a graph and optionally save.

        Args:
            1. graphPath (str, optional): The full path to write the graph to. Defaults to None.\n
            2. overwrite (bool, optional=True): Overwrite current graph values or not\n
            3. save (bool, optional=False): Save the graph to the self.graphPath or graphPath.\n
        """
        if graphPath and next(graphPath.endswith(extension) for extension in ut.RDF_EXTENSIONS) :
            self._graphPath=graphPath

        self._graph=Graph() 
        ut.bind_ontologies(self._graph)      
        nodeType=ut.get_node_type(str(type(self)))                
        self._graph.add((self.subject, RDF.type, nodeType ))  

        # enumerate attributes in node and write them to triples
        attributes = ut.get_variables_in_class(self)
        attributes = ut.clean_attributes_list(attributes)        
        pathlist = ut.get_paths_in_class(self)
                
        for attribute in attributes: 
            predicate = ut.match_uri(attribute)
            value=getattr(self,attribute)
            
            if value is not None:
                dataType=ut.get_data_type(value)
                temp=dataType.toPython()
                predtemp=predicate.toPython()

                if self._graph.value(self._subject, predicate, None)== str(value):
                    continue

                #check if exists
                elif overwrite:
                    self._graph.remove((self._subject, predicate, None))

                if 'linkedSubjects' in attribute:
                    if len(value) !=0:
                        value=[subject.toPython() for subject in self.linkedSubjects]
                    else:
                        continue
                
                elif attribute in pathlist:
                    if (self._graphPath):
                        folderPath=ut.get_folder_path(self.graphPath)
                        try:
                            value=os.path.relpath(value,folderPath)
                        except:
                            pass
                if 'string' not in dataType.toPython():        
                    self._graph.add((self._subject, predicate, Literal(value,datatype=dataType)))
                else:
                    self._graph.add((self._subject, predicate, Literal(value)))

        #Save graph
        if(save):
            self.save_graph(graphPath)            
        return self._graph

    def save_graph(self,graphPath : str = None) -> bool:
        """Serialize the graph in an RDF file on drive.
        The RDF graph will be stored in self.graphPath or provided graphPath (str).

        Args:
            graphPath (str, optional)\n

        Raises:
            ValueError: No valid graphPath if file/folder location is not found\n
            ValueError: No valid extension if not in ut.RDF_EXTENSIONS\n
            ValueError: Save failed despite valid graphPath and extension (serialization error).\n

        Returns:
            bool: True if file is succesfully saved.
        """
        #check path validity
        if(graphPath and ut.check_if_path_is_valid(graphPath)): 
            self._graphPath=graphPath
        elif ut.check_if_path_is_valid(self._graphPath):
            pass
        else: 
            raise ValueError('No valid graphPath.')
        #check extension
        if (ut.get_extension(graphPath) not in ut.RDF_EXTENSIONS):
            raise ValueError(''.join(ut.RDF_EXTENSIONS) + ' currently are only supported extensions.')

        try: 
            # f= open(self._graphPath, 'w') 
            # base=ut.get_folder(self.graphPath)
            self._graph.serialize(self._graphPath)#,base=base
            # f.close()
            if os.path.exists(self._graphPath):                
                return True

            return False
        except:
            raise ValueError('Save failed despite valid graphPath.') 

###############################################



"Help functions for generic combination algorithms"

import open3d as o3d
import matplotlib
from scipy.spatial import Delaunay
import numpy as np
import geomapi.tools.completiontools.params as params
from typing import List, Tuple

def hello_world():
    return "hello world!"

def get_geometry(path : str) -> o3d.geometry:
    """Gets a open3d Geometry from a path

    Args:
        path (str): The absulute path to the resource

    Returns:
        o3d.geometry: open3d.Pointcloud or open3d.Trianglemesh, depending on the extension
    """

    newGeometry = None
    if(path.endswith(tuple(params.MESH_EXTENSION))):
        newGeometry = o3d.io.read_triangle_mesh(path)
        #if not newGeometry.has_vertex_normals():
        newGeometry.compute_vertex_normals()
    elif(path.endswith(tuple(params.PCD_EXTENSION))):
        newGeometry = o3d.io.read_point_cloud(path)
        
    return newGeometry

def show_geometries(geometries : 'List[o3d.geometry]', color : bool = False):
    """Displays different types of geometry in a scene

    Args:
        geometries (List[open3d.geometry]): The list of geometries
        color (bool, optional): recolor the objects to have a unique color. Defaults to False.
    """

    viewer = o3d.visualization.Visualizer()
    viewer.create_window()
    frame = o3d.geometry.TriangleMesh.create_coordinate_frame()
    viewer.add_geometry(frame)
    for i, geometry in enumerate(geometries):
        if color:
            geometry.paint_uniform_color(matplotlib.colors.hsv_to_rgb([float(i)/len(geometries),0.8,0.8]))
        viewer.add_geometry(geometry)
    opt = viewer.get_render_option()
    opt.background_color = np.asarray([1,1,1])
    opt.light_on = True
    viewer.run()

def get_lineset(geometry: o3d.geometry.TriangleMesh, color: Tuple[float, float, float] = (1,0,0)) -> o3d.geometry.LineSet:
    """Returns a lineset representation of a mesh

    Args:
        geometry (open3d.geometry.trianglemesh): the mesh to convert
        color (Tuple, optional): the color to paint the lineset. Defaults to (1,0,0).

    Returns:
        open3d.geometry.LineSet: the lineset from the mesh
    """

    ls = o3d.geometry.LineSet.create_from_triangle_mesh(geometry)
    ls.paint_uniform_color(color)

    return ls

def get_indices_in_hull(points : np.array, hull :np.array) -> List[int]:
    """Get the indices of all tyhe points that are inside the hull

    Args:
        points (numpy.array): should be a `NxK` coordinates of `N` points in `K` dimensions
        hull (np.array): is either a scipy.spatial.Delaunay object or the `MxK` array of the 
coordinates of `M` points in `K`dimensions for which Delaunay triangulation
will be computed

    Returns:
        List[int]: the indices of the points that are in the hull
    """

    if not isinstance(hull,Delaunay):
        hull = Delaunay(hull)
    ind = hull.find_simplex(points)>=0

    intList = []
    for i,x in enumerate(ind):
        if ind[i]:
            intList.append(i)

    return intList

def filter_pcd_by_distance(sourcePcd : o3d.geometry.PointCloud, testPcd: o3d.geometry.PointCloud, maxDistance : float) -> Tuple[o3d.geometry.PointCloud,o3d.geometry.PointCloud]:
    """Splits the sourcePcd in close and too far point compared to the testpcd

    Args:
        sourcePcd (open3d.geometry.PointCloud): The pcd to be split
        o3d (open3d.geometry.PointCloud): the pcd to test against

    Returns:
        Tuple(o3d.geometry.PointCloud,o3d.geometry.PointCloud): The points close enough and the points too far away
    """

    dists = sourcePcd.compute_point_cloud_distance(testPcd)
    dists = np.asarray(dists)
    ind = np.where(dists < maxDistance)[0]

    return (sourcePcd.select_by_index(ind), sourcePcd.select_by_index(ind, True))


"""A package to Manage and Work with RDF based Geomatic data"""

from .node import Node
from .geometrynode import GeometryNode
from .pointcloudnode import PointCloudNode
from .bimnode import BIMNode
from .meshnode import MeshNode
from .imagenode import ImageNode
from .orthonode import OrthoNode
from .linesetnode import LinesetNode
from .sessionnode import SessionNode





import copy
import math
import os
import shutil
import time
import unittest
from multiprocessing.sharedctypes import Value

import cv2
import geomapi.utils as ut
import geomapi.utils.geometryutils as gt
import ifcopenshell
import numpy as np
import open3d as o3d
import PIL
import pye57
import rdflib
from geomapi.nodes import *
from ifcopenshell.util.selector import Selector
from PIL import Image
from rdflib import RDF, RDFS, Graph, Literal, URIRef


class TestNode(unittest.TestCase):

################################## SETUP/TEARDOWN CLASS ######################
    @classmethod
    def setUpClass(cls):
        #execute once before all tests
        print('-----------------Setup Class----------------------')
        st = time.time()
        cls.path=os.path.join(os.getcwd(), "test","testfiles") 

        #ONTOLOGIES
        cls.exif = rdflib.Namespace('http://www.w3.org/2003/12/exif/ns#')
        cls.geo=rdflib.Namespace('http://www.opengis.net/ont/geosparql#') #coordinate system information
        cls.gom=rdflib.Namespace('https://w3id.org/gom#') # geometry representations => this is from mathias
        cls.omg=rdflib.Namespace('https://w3id.org/omg#') # geometry relations
        cls.fog=rdflib.Namespace('https://w3id.org/fog#')
        cls.v4d=rdflib.Namespace('https://w3id.org/v4d/core#')
        cls.openlabel=rdflib.Namespace('https://www.asam.net/index.php?eID=dumpFile&t=f&f=3876&token=413e8c85031ae64cc35cf42d0768627514868b2f#')
        cls.e57=rdflib.Namespace('http://libe57.org#')
        cls.xcr=rdflib.Namespace('http://www.w3.org/1999/02/22-rdf-syntax-ns#')
        cls.ifc=rdflib.Namespace('http://ifcowl.openbimstandards.org/IFC2X3_Final#')
       
        #IMGGRAPH
        cls.imgGraphPath=os.path.join(cls.path,'imgGraph.ttl')
        cls.imgGraph=Graph().parse(cls.imgGraphPath)

        #IMG1 (image)
        cls.path1=os.path.join(cls.path,'IMG',"DJI_0067.JPG")          
        cls.img1=cv2. imread(cls.path1)          
        cls.subject1=next(s for s in cls.imgGraph.subjects(RDF.type) if "file:///DJI_0067" in s.toPython())
        
        #IMG2 (PIL Image)
        cls.path2=os.path.join(cls.path,'IMG',"DJI_0070.JPG")          
        cls.img2=Image.open( cls.path2)     
        cls.subject2=next(s for s in cls.imgGraph.subjects(RDF.type) if "file:///DJI_0070" in s.toPython())
        
        #IMG3 (XMP file)
        cls.path3=os.path.join(cls.path,'IMG',"IMG_2173.JPG")   
        cls.xmpPath3= os.path.join(cls.path,'IMG',"IMG_2173.xmp")     
        cls.img3=Image.open( cls.path3)     
        cls.subject3=next(s for s in cls.imgGraph.subjects(RDF.type) if "file:///IMG_2173" in s.toPython())
         
        #IMG3 (XML large file)
        cls.path4=os.path.join(cls.path,'IMG',"101_0366_0036.jpg")   
        cls.xmlPath4= os.path.join(cls.path,'IMG',"ReferenceLAMBERT08_TAW.xml")     
        cls.img4=Image.open( cls.path4)     
        cls.subject4=next(s for s in cls.imgGraph.subjects(RDF.type) if "file:///101_0366_0036" in s.toPython())
        
        #IMG5 (XML large file)
        cls.path5=os.path.join(cls.path,'IMG',"IMG_202200706_00_ (1).jpg")   
        cls.xmlPath5= os.path.join(cls.path,'IMG',"camera_position.xml")     
        cls.img5=Image.open( cls.path5)
        cls.subject5=next(s for s in cls.imgGraph.subjects(RDF.type) if "file:///IMG_202200706_00___1_" in s.toPython())

        #RESOURCES
        cls.resourcePath=os.path.join(cls.path,"resources")
        if not os.path.exists(cls.resourcePath):
            os.mkdir(cls.resourcePath)

        #FILES
        cls.files=ut.get_list_of_files(cls.path)

        #TIME TRACKING           
        et = time.time()
        print("startup time: "+str(et - st))
        print('{:50s} {:5s} '.format('tests','time'))
        print('------------------------------------------------------')

    @classmethod
    def tearDownClass(cls):
        #execute once after all tests
        print('-----------------TearDown Class----------------------')
        if os.path.exists(cls.resourcePath):
            shutil.rmtree(cls.resourcePath)  
        if os.path.exists(os.path.join(os.getcwd(),'IMG') ):
            shutil.rmtree(os.path.join(os.getcwd(),'IMG') )
        if os.path.exists(os.path.join(os.getcwd(),'resources') ):
            shutil.rmtree(os.path.join(os.getcwd(),'resources') )

################################## SETUP/TEARDOWN ######################

    def setUp(self):
        #execute before every test
        self.startTime = time.time()   

    def tearDown(self):
        #execute after every test
        t = time.time() - self.startTime
        print('{:50s} {:5s} '.format(self._testMethodName,str(t)))
################################## TEST FUNCTIONS ######################

    def test_ImageNode_creation_from_subject(self):
        #subject
        subject='myNode'
        node= ImageNode(subject=subject)
        self.assertEqual(node.subject.toPython(),'file:///'+subject)

        #http://
        subject='http://session_2022_05_20'
        node= ImageNode(subject=subject)
        self.assertEqual(node.subject.toPython(),subject)
        
        #erroneous char       
        subject='[[http://ses>sion_2022_<05_20]]'
        node= ImageNode(subject=subject)
        self.assertEqual(node.subject.toPython(),'http://__ses_sion_2022__05_20__')

    def test_ImageNode_creation_from_graph(self):
        subject=next(self.imgGraph.subjects(RDF.type))
        node= ImageNode(graph=self.imgGraph, subject=subject)
        self.assertEqual(node.subject,subject)
        
    def test_ImageNode_creation_from_graph_path(self):
        subject=next(self.imgGraph.subjects(RDF.type))
        node= ImageNode(graphPath=self.imgGraphPath, subject=subject)
        self.assertEqual(node.subject,subject)
        
    def test_ImageNode_creation_from_path(self):
        #path1 without getResource
        node= ImageNode(path=self.path1)
        self.assertEqual(node.name,ut.get_filename(self.path1))
        self.assertIsNone(node.resource)

        #path2 with getResource
        node= ImageNode(path=self.path2,getResource=True)        
        self.assertEqual(node.name,ut.get_filename(self.path2))
        self.assertEqual(node.imageHeight,self.img2.height)

        #path3 
        node= ImageNode(path=self.path3,getResource=True)
        self.assertEqual(node.name,ut.get_filename(self.path3))
        self.assertEqual(node.imageHeight,self.img3.height)

    def test_ImageNode_creation_from_xmpPath(self):
        #path without extra info 
        node= ImageNode(xmpPath=self.xmpPath3)
        self.assertIsNotNone(node.name)

        #path with getResource without extra info 
        node= ImageNode(xmpPath=self.xmpPath3,getResource=True)        
        self.assertIsNotNone(node.name)
        self.assertIsNotNone(node.resource)


    def test_ImageNode_creation_from_xmlPath(self):
        
        
        #path without extra info 
        node= ImageNode(xmlPath=self.xmlPath4,subject='101_0366_0037')
        self.assertIsNotNone(node.name)

        #path with subject
        node= ImageNode(xmlPath=self.xmlPath4,subject='101_0366_0037',getResource=True)        
        self.assertIsNotNone(node.name)
        self.assertIsNotNone(node.resource)

    def test_ImageNode_creation_from_resource(self):
        #img1
        node= ImageNode(resource=self.img1)
        self.assertEqual(node.imageHeight,self.img1.shape[0])
        #img2
        node= ImageNode(resource=self.img2)
        self.assertEqual(node.imageHeight,self.img1.shape[0])
        #img3
        node= ImageNode(resource=self.img3)
        self.assertEqual(node.imageHeight,self.img3.height)

    def test_creation_from_subject_and_graph_and_graphPath(self):        
        subject=next(self.imgGraph.subjects(RDF.type))
        node= ImageNode(subject=subject,graph=self.imgGraph,graphPath=self.imgGraphPath)
        self.assertEqual(node.subject,subject)
        
    def test_creation_from_subject_and_path(self):        
        node= ImageNode(subject=self.subject2,path=self.path2,getResource=True)
        self.assertEqual(node.subject,self.subject2)
        node.to_graph()
        self.assertEqual(node.imageHeight,self.img2.height)
        
    def test_creation_from_subject_and_path_and_graph(self):      
        node= ImageNode(subject=self.subject3,path=self.path3,graph=self.imgGraph,getResource=True)
        self.assertEqual(node.subject,self.subject3)
        node.to_graph()
        initialGraph=ut.get_subject_graph(self.imgGraph,subject=self.subject3)
        self.assertEqual(len(node.graph),len(initialGraph))

    def test_creation_from_resource_and_path (self):      
        node= ImageNode(resource=self.img5,path=self.path5)
        self.assertEqual(node.subject,self.subject5)

    def test_creation_from_subject_and_resource_and_path_and_graph(self):      
        node= ImageNode(subject=self.subject3,resource=self.img3,path=self.path3, graph=self.imgGraph)
        self.assertEqual(node.subject,self.subject3)
        node.to_graph()
        object=node.graph.value(node.subject,self.v4d['path'])
        self.assertEqual(object.toPython(),os.path.join("IMG",ut.get_filename(self.path3)+'.JPG') )

    def test_creation_from_subject_resource_path_xmpPath_graph(self):      
        node= ImageNode(subject=self.subject3,resource=self.img3,path=self.path3, xmpPath=self.xmpPath3, graph=self.imgGraph)
        self.assertEqual(node.subject,self.subject3)
        node.to_graph()
        object=node.graph.value(node.subject,self.v4d['path'])
        self.assertEqual(object.toPython(),os.path.join(self.path,"IMG",ut.get_filename(self.path3)+'.JPG') )

    def test_node_creation_with_get_resource(self):
        #mesh
        node= ImageNode(resource=self.img1)
        self.assertIsNotNone(node.resource)

        #path without getResource
        node= ImageNode(path=self.path2)
        self.assertIsNone(node.resource)

        #path with getResource
        node= ImageNode(path=self.path5,getResource=True)
        self.assertIsNotNone(node.resource)

        #graph with get resource
        node= ImageNode(subject=self.subject2,graph=self.imgGraph,getResource=True)
        self.assertIsNone(node.resource)
        
        #graphPath with get resource
        node= ImageNode(subject=self.subject3,graphPath=self.imgGraphPath,getResource=True)
        self.assertIsNotNone(node.resource)

    def test_clear_resource(self):
        #mesh
        node= ImageNode(resource=self.img1)
        self.assertIsNotNone(node.resource)
        del node.resource
        self.assertIsNone(node.resource)

    def test_save_resource(self):
        #no mesh -> False
        node= ImageNode()
        self.assertFalse(node.save_resource())

        #directory
        node= ImageNode(resource=self.img2)
        self.assertTrue(node.save_resource(os.path.join(self.path,'resources')))

        #graphPath        
        node= ImageNode(resource=self.img2,graphPath=self.imgGraphPath)
        self.assertTrue(node.save_resource())

        #no path or graphPath
        node= ImageNode(resource=self.img4)        
        self.assertTrue(node.save_resource())

        #path -> new name
        node= ImageNode(subject=URIRef('myImg'),path=self.path2,getResource=True)
        self.assertTrue(node.save_resource())
        self.assertTrue( node.path in self.files)
        
        #graphPath with directory
        node=ImageNode(subject=self.subject5,graphPath=self.imgGraphPath, resource=self.img5)
        self.assertTrue(node.save_resource(os.path.join(self.path,'resources')))

        #graph with new subject
        node=ImageNode(subject=self.subject4,graph=self.imgGraph, resource=self.img4)
        node.subject='myImg'
        self.assertTrue(node.save_resource())

    def test_get_resource(self):
        #mesh
        node=ImageNode(resource=self.img3)  
        self.assertIsNotNone(node.get_resource())

        #no mesh
        del node.resource
        self.assertIsNone(node.get_resource())

        #graphPath with getResource
        node=ImageNode(graphPath=self.imgGraphPath,subject=self.subject2,getResource=True)
        self.assertIsNotNone(node.get_resource())

    def test_set_path(self):
        #valid path
        node=ImageNode()
        node.path= self.path1
        self.assertEqual(node.path,self.path1)

        #preexisting
        node=ImageNode(path=self.path4)
        self.assertEqual(node.path,self.path4)

        #graphPath & name
        node=ImageNode(subject=self.subject5,graphPath=self.imgGraphPath)
        node.get_path()
        self.assertEqual(node.path,self.path5)

    def test_set_xmp_path(self):
        #valid path
        node=ImageNode(xmpPath=self.xmpPath3)
        self.assertEqual(node.xmpPath,self.xmpPath3)

        #invalid
        self.assertRaises(ValueError,ImageNode,xmpPath='qsffqsdf.dwg')

    def test_set_xml_path(self):
        #valid path
        node=ImageNode()
        node.xmlPath=self.xmlPath4
        self.assertEqual(node.xmlPath,self.xmlPath4)

        #invalid
        self.assertRaises(ValueError,ImageNode,xmlPath='qsffqsdf.dwg')

if __name__ == '__main__':
    unittest.main()
